/**
 * @file main.c
 * @brief Main routine
 *
 * @section License
 *
 * Copyright (C) 2010-2018 Oryx Embedded SARL. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * @author Oryx Embedded SARL (www.oryx-embedded.com)
 * @version 1.9.0
 **/

//Dependencies
#include <stdlib.h>
#include <string.h>
#include "stm32f4xx.h"
#include "stm32f4xx_hal.h"
#include "stm32f4xx_hal_rcc.h"
#include "stm32f4xx_hal_pwr.h"
#include "stm32f4xx_hal_pwr_ex.h"
#include "stm32f4xx_hal_flash.h"
#include "stm32f4xx_ll_fsmc.h"
#include "cmsis_os.h"
#include "Board_LED.h"
#include "Board_Buttons.h"
#include "Board_GLCD.h"
#include "GLCD_Config.h"

//CMSIS-RTOS and network stack dependencies
#include "cmsis_os.h"
#include "rl_net.h"

//SSL library dependencies
#include "tls.h"
#include "tls_cipher_suites.h"
#include "tls_misc.h"
#include "rng/yarrow.h"
#include "resource_manager.h"
#include "error.h"
#include "debug.h"

//Server port
#define APP_SERVER_PORT 443

//Diffie-Hellman parameters
#define APP_SERVER_DH_PARAMS "certs/dh_params.pem"

//Server's RSA certificate and private key
#define APP_SERVER_RSA_CERT "certs/server_rsa_cert.pem"
#define APP_SERVER_RSA_PRIVATE_KEY "certs/server_rsa_key.pem"

//Server's ECDSA certificate and private key
#define APP_SERVER_ECDSA_CERT "certs/server_ecdsa_cert.pem"
#define APP_SERVER_ECDSA_PRIVATE_KEY "certs/server_ecdsa_key.pem"

//Trusted CA bundle
#define APP_CA_CERT_BUNDLE "certs/ca_cert_bundle.pem"

//Credentials
char_t *dhParams = NULL;
size_t dhParamsLength = 0;
char_t *serverRsaCert = NULL;
size_t serverRsaCertLength = 0;
char_t *serverRsaPrivateKey = NULL;
size_t serverRsaPrivateKeyLength = 0;
char_t *serverEcdsaCert = NULL;
size_t serverEcdsaCertLength = 0;
char_t *serverEcdsaPrivateKey = NULL;
size_t serverEcdsaPrivateKeyLength = 0;
char_t *trustedCaList = NULL;
size_t trustedCaListLength = 0;

//Pseudo-random number generator
YarrowContext yarrowContext;

//Session cache
TlsCache *tlsCache = NULL;

//Global variables
uint_t lcdLine = 0;
uint_t lcdColumn = 0;
uint_t hitCounter = 0;

RNG_HandleTypeDef RNG_Handle;

YarrowContext yarrowContext;
uint8_t seed[32];

//Thread definitions
void sslServerThread(void const *arg);
osThreadDef(sslServerThread, osPriorityNormal, 1, 1000);

void userThread(void const *arg);
osThreadDef(userThread, osPriorityNormal, 1, 1000);

void blinkThread(void const *arg);
osThreadDef(blinkThread, osPriorityNormal, 1, 200);

//RTOS object definitions
osSemaphoreId connectionSemaphore;
osSemaphoreDef(connectionSemaphore);

//External variables
extern GLCD_FONT GLCD_Font_16x24;

//Forward declaration of functions
void sslClientThread(void const *arg);

const char_t *dumpArray(char_t *buffer, const uint8_t *data, size_t length);

error_t tlsSendCallback(TlsIoHandle socket, const void *data,
   size_t length, size_t *written, uint_t flags);

error_t tlsReceiveCallback(TlsIoHandle socket, void *data,
   size_t size, size_t *received, uint_t flags);


/**
 * @brief Set cursor location
 * @param[in] line Line number
 * @param[in] column Column number
 **/

void lcdSetCursor(uint_t line, uint_t column)
{
   lcdLine = MIN(line, 10);
   lcdColumn = MIN(column, 20);
}


/**
 * @brief Write a character to the LCD display
 * @param[in] c Character to be written
 **/

void lcdPutChar(char_t c)
{
   if(c == '\r')
   {
      lcdColumn = 0;
   }
   else if(c == '\n')
   {
      lcdColumn = 0;
      lcdLine++;
   }
   else if(lcdLine < 10 && lcdColumn < 20)
   {
      //Display current character
      GLCD_DrawChar(lcdColumn * 16, lcdLine * 24, c);

      //Advance the cursor position
      if(++lcdColumn >= 20)
      {
         lcdColumn = 0;
         lcdLine++;
      }
   }
}


/**
 * @brief System clock configuration
 **/

void SystemClock_Config(void)
{
   RCC_ClkInitTypeDef RCC_ClkInitStruct;
   RCC_OscInitTypeDef RCC_OscInitStruct;

   //Enable Power Control clock
   __HAL_RCC_PWR_CLK_ENABLE();

   //The voltage scaling allows optimizing the power consumption when the device is
   //clocked below the maximum system frequency, to update the voltage scaling value
   //regarding system frequency refer to product datasheet.
   __HAL_PWR_VOLTAGESCALING_CONFIG(PWR_REGULATOR_VOLTAGE_SCALE1);

   //Enable HSE Oscillator and activate PLL with HSE as source
   RCC_OscInitStruct.OscillatorType = RCC_OSCILLATORTYPE_HSE;
   RCC_OscInitStruct.HSEState = RCC_HSE_ON;
   RCC_OscInitStruct.PLL.PLLState = RCC_PLL_ON;
   RCC_OscInitStruct.PLL.PLLSource = RCC_PLLSOURCE_HSE;
   RCC_OscInitStruct.PLL.PLLM = 25;
   RCC_OscInitStruct.PLL.PLLN = 336;
   RCC_OscInitStruct.PLL.PLLP = RCC_PLLP_DIV2;
   RCC_OscInitStruct.PLL.PLLQ = 7;
   HAL_RCC_OscConfig(&RCC_OscInitStruct);

   //HAL_PWREx_ActivateOverDrive();

   //Select PLL as system clock source and configure the HCLK, PCLK1 and PCLK2
   //clocks dividers
   RCC_ClkInitStruct.ClockType = (RCC_CLOCKTYPE_SYSCLK | RCC_CLOCKTYPE_HCLK | RCC_CLOCKTYPE_PCLK1 | RCC_CLOCKTYPE_PCLK2);
   RCC_ClkInitStruct.SYSCLKSource = RCC_SYSCLKSOURCE_PLLCLK;
   RCC_ClkInitStruct.AHBCLKDivider = RCC_SYSCLK_DIV1;
   RCC_ClkInitStruct.APB1CLKDivider = RCC_HCLK_DIV4;
   RCC_ClkInitStruct.APB2CLKDivider = RCC_HCLK_DIV2;
   HAL_RCC_ClockConfig(&RCC_ClkInitStruct, FLASH_LATENCY_5);
}


/**
 * @brief User thread
 **/

void userThread(void const *arg)
{
   uint8_t ipv4Addr[NET_ADDR_IP4_LEN];
   char_t buffer[20];

   lcdSetCursor(2, 0);
   printf("IPv4 Addr\r\n");
   lcdSetCursor(5, 0);
   printf("Subnet Mask\r\n");

   //Endless loop
   while(1)
   {
      //Move cursor
      lcdSetCursor(3, 0);

      //Retrieve IPv4 address
      netIF_GetOption(NET_IF_CLASS_ETH, netIF_OptionIP4_Address,
         ipv4Addr, NET_ADDR_IP4_LEN);

      //Display IPv4 address
      netIP_ntoa(NET_ADDR_IP4, ipv4Addr, buffer, sizeof(buffer));
      printf("%s        \r\n", buffer);

      //Move cursor
      lcdSetCursor(6, 0);

      //Retrieve subnet mask
      netIF_GetOption(NET_IF_CLASS_ETH, netIF_OptionIP4_SubnetMask,
         ipv4Addr, NET_ADDR_IP4_LEN);

      //Display subnet mask
      netIP_ntoa(NET_ADDR_IP4, ipv4Addr, buffer, sizeof(buffer));
      printf("%s        \r\n", buffer);

      //Delay
      osDelay(1000);
   }
}


/**
 * @brief LED blinking thread
 **/

void blinkThread(void const *arg)
{
   //Endless loop
   while(1)
   {
      LED_On(0);
      osDelay(100);
      LED_Off(0);
      osDelay(900);
   }
}


/**
 * @brief Main entry point
 * @return Unused value
 **/

int main(void)
{
   osThreadId id;
   netStatus status;

   //HAL library initialization
   HAL_Init();
   //System clock configuration
   SystemClock_Config();

   //Initialize kernel
   osKernelInitialize();
   //Configure debug UART
   debugInit(115200);

   //Start-up message
   TRACE_INFO("\r\n");
   TRACE_INFO("******************************\r\n");
   TRACE_INFO("*** CycloneSSL Server Demo ***\r\n");
   TRACE_INFO("******************************\r\n");
   TRACE_INFO("Copyright: 2010-2018 Oryx Embedded SARL\r\n");
   TRACE_INFO("Compiled: %s %s\r\n", __DATE__, __TIME__);
   TRACE_INFO("Target: STM32F407\r\n");
   TRACE_INFO("\r\n");

   //LED configuration
   LED_Initialize();
   //Clear LEDs
   LED_Off(0);
   LED_Off(1);
   LED_Off(2);
   LED_Off(3);
   LED_Off(4);
   LED_Off(5);
   LED_Off(6);
   LED_Off(7);

   //Initialize buttons
   Buttons_Initialize();

   //Initialize graphic LCD display
   GLCD_Initialize();
   GLCD_SetForegroundColor(GLCD_COLOR_WHITE);
   GLCD_SetBackgroundColor(GLCD_COLOR_BLUE);
   GLCD_ClearScreen();
   GLCD_SetFont(&GLCD_Font_16x24);

   //Welcome message
   lcdSetCursor(0, 0);
   printf("SSL Server Demo");

   //Enable RNG peripheral clock
   __HAL_RCC_RNG_CLK_ENABLE();
   //Initialize RNG
   RNG_Handle.Instance = RNG;
   HAL_RNG_Init(&RNG_Handle);

   //TCP/IP stack initialization
   status = netInitialize();
   //Any error to report?
   if(status != netOK)
   {
      //Debug message
      TRACE_ERROR("Failed to initialize TCP/IP stack!\r\n");
   }

   //Create a thread to handle incoming requests
   id = osThreadCreate(osThread(sslServerThread), NULL);
   //Failed to create the task?
   if(id == NULL)
   {
      //Debug message
      TRACE_ERROR("Failed to create thread!\r\n");
   }

   //Create the user thread
   id = osThreadCreate(osThread(userThread), NULL);
   //Failed to create the thread?
   if(id == NULL)
   {
      //Debug message
      TRACE_ERROR("Failed to create thread!\r\n");
   }

   //Create a thread to blink the LED
   id = osThreadCreate(osThread(blinkThread), NULL);
   //Failed to create the thread?
   if(id == NULL)
   {
      //Debug message
      TRACE_ERROR("Failed to create thread!\r\n");
   }

   //Start the kernel
   osKernelStart();

   //This function should never return
   return 0;
}


/**
 * @brief SSL server task
 * @param param[in] arg Not used
 **/

void sslServerThread(const void *arg)
{
   error_t error;
   uint_t i;
   uint_t counter;
   uint32_t value;
   int ret;
   int serverSocket;
   int clientSocket;
   int clientAddrLen;
   SOCKADDR_IN serverAddr;
   SOCKADDR_IN clientAddr;
   osThreadDef_t threadDef;
   osThreadId id;
   uint8_t seed[32];

   //Create a semaphore to limit the number of simultaneous connections
   connectionSemaphore = osSemaphoreCreate(osSemaphore(connectionSemaphore), 4);
   //Any error to report?
   if(connectionSemaphore == NULL)
   {
      //Debug message
      TRACE_ERROR("Failed to create semaphore!\r\n");
   }

   //Generate a random seed
   for(i = 0; i < 32; i += 4)
   {
      //Get 32-bit random value
      HAL_RNG_GenerateRandomNumber(&RNG_Handle, &value);

      //Copy random value
      seed[i] = value & 0xFF;
      seed[i + 1] = (value >> 8) & 0xFF;
      seed[i + 2] = (value >> 16) & 0xFF;
      seed[i + 3] = (value >> 24) & 0xFF;
   }

   //PRNG initialization
   error = yarrowInit(&yarrowContext);
   //Any error to report?
   if(error)
   {
      //Debug message
      TRACE_ERROR("Failed to initialize PRNG!\r\n");
   }

   //Properly seed the PRNG
   error = yarrowSeed(&yarrowContext, seed, sizeof(seed));
   //Any error to report?
   if(error)
   {
      //Debug message
      TRACE_ERROR("Failed to seed PRNG!\r\n");
   }

   //Start of exception handling block
   do
   {
      //Load Diffie-Hellman parameters
      error = resGetData(APP_SERVER_DH_PARAMS, (uint8_t **) &dhParams, &dhParamsLength);
      //Any error to report?
      if(error)
         break;

      //Load server's RSA certificate
      error = resGetData(APP_SERVER_RSA_CERT, (uint8_t **) &serverRsaCert, &serverRsaCertLength);
      //Any error to report?
      if(error)
         break;

      //Load server's RSA private key
      error = resGetData(APP_SERVER_RSA_PRIVATE_KEY, (uint8_t **) &serverRsaPrivateKey, &serverRsaPrivateKeyLength);
      //Any error to report?
      if(error)
         break;

      //Load server's ECDSA certificate
      error = resGetData(APP_SERVER_ECDSA_CERT, (uint8_t **) &serverEcdsaCert, &serverEcdsaCertLength);
      //Any error to report?
      if(error)
         break;

      //Load server's ECDSA private key
      error = resGetData(APP_SERVER_ECDSA_PRIVATE_KEY, (uint8_t **) &serverEcdsaPrivateKey, &serverEcdsaPrivateKeyLength);
      //Any error to report?
      if(error)
         break;

      //Load trusted CA certificates
      error = resGetData(APP_CA_CERT_BUNDLE, (uint8_t **) &trustedCaList, &trustedCaListLength);
      //Any error to report?
      if(error)
         break;

      //End of exception handling block
   } while(0);

   //Check error code
   if(error)
   {
      //Debug message
      TRACE_ERROR("Failed to load credentials!\r\n");
   }

   //TLS session cache initialization
   tlsCache = tlsInitCache(16);
   //Any error to report?
   if(!tlsCache)
   {
      //Debug message
      TRACE_ERROR("TLS cache initialization failed!\r\n");
   }

   //Open a socket
   serverSocket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
   //Failed to open socket?
   if(serverSocket < 0)
   {
      //Debug message
      TRACE_ERROR("Error: Cannot open socket\r\n");
   }

   //Server port (443)
   serverAddr.sin_port = htons(APP_SERVER_PORT);
   serverAddr.sin_family = PF_INET;
   serverAddr.sin_addr.s_addr = INADDR_ANY;

   //Bind the socket to the relevant port number
   ret = bind(serverSocket, (SOCKADDR *) &serverAddr, sizeof(serverAddr));
   //Any error to report?
   if(ret < 0)
   {
      //Debug message
      TRACE_ERROR("Error: Failed to bind socket\r\n");
   }

   //Place the socket in the listening state
   ret = listen(serverSocket, 1);
   //Any error to report?
   if(ret < 0)
   {
      //Debug message
      TRACE_ERROR("Error: Failed to enter listening state\r\n");
   }

   //Process incoming connections to the server
   for(counter = 1; ; counter++)
   {
      //Debug message
      TRACE_INFO("\r\n\r\n");
      TRACE_INFO("Waiting for an incoming connection...\r\n\r\n");

      //Initialize address length
      clientAddrLen = sizeof(clientAddr);

      //Accept a new connection request
      clientSocket = accept(serverSocket, (SOCKADDR *) &clientAddr, &clientAddrLen);

      //Any error to report?
      if(clientSocket < 0)
      {
         //Debug message
         TRACE_ERROR("Error: Cannot accept new connection request\r\n");
         continue;
      }

      //Limit the number of simultaneous connections to the server
      if(!osSemaphoreWait(connectionSemaphore, 0))
      {
         //Debug message
         TRACE_INFO("Connection #%u refused with client %u.%u.%u.%u port %u...\r\n",
            counter, clientAddr.sin_addr.s_b1, clientAddr.sin_addr.s_b2,
            clientAddr.sin_addr.s_b3, clientAddr.sin_addr.s_b4,
            ntohs(clientAddr.sin_port));
         //Close socket
         closesocket(clientSocket);
         //Connection request is refused
         continue;
      }

      //Increment hit counter
      hitCounter++;

      //Debug message
      TRACE_INFO("Connection #%u established with client %u.%u.%u.%u port %u...\r\n",
         hitCounter, clientAddr.sin_addr.s_b1, clientAddr.sin_addr.s_b2,
         clientAddr.sin_addr.s_b3, clientAddr.sin_addr.s_b4,
         ntohs(clientAddr.sin_port));

      //thread parameters
      threadDef.pthread = sslClientThread;
      threadDef.tpriority = osPriorityNormal;
      threadDef.instances = 1;
      threadDef.stacksize = 1000;

      //Create a thread to service the client connection
      id = osThreadCreate(&threadDef, (void *) clientSocket);

      //Did we encounter an error?
      if(id == NULL)
      {
         //Debug message
         TRACE_ERROR("Failed to create task!\r\n");
         //Close socket
         closesocket(clientSocket);
         //Release semaphore
         osSemaphoreRelease(connectionSemaphore);
      }
   }
}


/**
 * @brief SSL client task
 * @param param[in] arg Client socket
 **/

void sslClientThread(const void *arg)
{
   error_t error;
   size_t n;
   int clientSocket;
   char *p;
   TlsContext *tlsContext = NULL;
   char *buffer = NULL;
   char *response = NULL;

   //Retrieve socket handle
   clientSocket = (int) arg;

   //Start of exception handling block
   do
   {
      //Allocate a memory buffer
      buffer = malloc(512);
      //Failed to allocate memory?
      if(buffer == NULL)
         break;

      //Allocate a memory buffer to hold the response
      response = malloc(2048);
      //Failed to allocate memory?
      if(response == NULL)
         break;

      //Initialize SSL/TLS context
      tlsContext = tlsInit();
      //Initialization failed?
      if(!tlsContext)
      {
         //Report an error
         error = ERROR_OUT_OF_MEMORY;
         //Exit immediately
         break;
      }

      //Select server operation mode
      error = tlsSetConnectionEnd(tlsContext, TLS_CONNECTION_END_SERVER);
      //Any error to report?
      if(error)
         break;

      //Set send and receive callbacks (I/O abstraction layer)
      error = tlsSetIoCallbacks(tlsContext, (TlsIoHandle) clientSocket,
         tlsSendCallback, tlsReceiveCallback);
      //Any error to report?
      if(error)
         break;

      //Set the PRNG algorithm to be used
      error = tlsSetPrng(tlsContext, YARROW_PRNG_ALGO, &yarrowContext);
      //Any error to report?
      if(error)
         break;

      //Session cache that will be used to save/resume TLS sessions
      error = tlsSetCache(tlsContext, tlsCache);
      //Any error to report?
      if(error)
         break;

      //Client authentication is optional
      error = tlsSetClientAuthMode(tlsContext, TLS_CLIENT_AUTH_NONE);
      //Any error to report?
      if(error)
         break;

#if (TLS_DHE_RSA_SUPPORT == ENABLED || TLS_DHE_DSS_SUPPORT == ENABLED || TLS_DH_ANON_SUPPORT == ENABLED)
      //Import Diffie-Hellman parameters
      error = tlsSetDhParameters(tlsContext, dhParams, dhParamsLength);
      //Any error to report?
      if(error)
         break;
#endif

#if (TLS_RSA_SUPPORT == ENABLED || TLS_DHE_RSA_SUPPORT == ENABLED || TLS_ECDHE_RSA_SUPPORT == ENABLED)
      //Import the server's RSA certificate
      error = tlsAddCertificate(tlsContext, serverRsaCert,
         serverRsaCertLength, serverRsaPrivateKey, serverRsaPrivateKeyLength);
      //Any error to report?
      if(error)
         break;
#endif

#if (TLS_ECDHE_ECDSA_SUPPORT == ENABLED)
      //Import the server's ECDSA certificate
      error = tlsAddCertificate(tlsContext, serverEcdsaCert,
         serverEcdsaCertLength, serverEcdsaPrivateKey, serverEcdsaPrivateKeyLength);
      //Any error to report?
      if(error)
         break;
#endif

      //Import the list of trusted CA certificates
      error = tlsSetTrustedCaList(tlsContext, trustedCaList, trustedCaListLength);
      //Any error to report?
      if(error)
         break;

      //Establish a secure session
      error = tlsConnect(tlsContext);
      //TLS handshake failure?
      if(error)
         break;

      //Debug message
      TRACE_INFO("\r\n");
      TRACE_INFO("HTTP request:\r\n");

      //Read HTTP request
      while(1)
      {
         //Read a complete line
         error = tlsRead(tlsContext, buffer, 512, &n, TLS_FLAG_BREAK_CRLF);
         //Any error to report?
         if(error)
            break;

         //Properly terminate the string with a NULL character
         buffer[n] = '\0';
         //Dump HTTP request
         TRACE_INFO("%s", buffer);

         //The end of the header has been reached?
         if(!strcmp(buffer, "\r\n")) break;
      }

      //Propagate exception if necessary...
      if(error)
         break;

      //Point to the beginning of the response
      p = response;

      //Format response
      p += sprintf(p, "HTTP/1.0 200 OK\r\n");
      p += sprintf(p, "Content-Type: text/html\r\n");
      p += sprintf(p, "\r\n");

      p += sprintf(p, "<!DOCTYPE html>\r\n");
      p += sprintf(p, "<html>\r\n");
      p += sprintf(p, "<head>\r\n");
      p += sprintf(p, "  <title>Oryx Embedded - CycloneSSL Server Demo</title>\r\n");
      p += sprintf(p, "  <style>\r\n");
      p += sprintf(p, "    body {font-family: monospace; font-size: 13px;}\r\n");
      p += sprintf(p, "    table {border-width: 1px; border-style: ouset; border-collapse: collapse;}\r\n");
      p += sprintf(p, "    td {border-width: 1px; border-style: inset; padding: 3px;}\r\n");
      p += sprintf(p, "  </style>\r\n");
      p += sprintf(p, "</head>\r\n");
      p += sprintf(p, "<body>\r\n");
      p += sprintf(p, "  <p>Welcome to the CycloneSSL demo server!</p>\r\n");
      p += sprintf(p, "  <table>\r\n");

      p += sprintf(p, "  <tr>\r\n");
      p += sprintf(p, "    <td>Hit counter</td>\r\n");
      p += sprintf(p, "    <td>%d</td>\r\n", hitCounter);
      p += sprintf(p, "  </tr>\r\n");

      p += sprintf(p, "  <tr>\r\n");
      p += sprintf(p, "    <td>Server version</td>\r\n");
      p += sprintf(p, "    <td>%s</td>\r\n", tlsGetVersionName(TLS_MAX_VERSION));
      p += sprintf(p, "  </tr>\r\n");

      p += sprintf(p, "  <tr>\r\n");
      p += sprintf(p, "    <td>Client version</td>\r\n");
      p += sprintf(p, "    <td>%s</td>\r\n", tlsGetVersionName(tlsContext->clientVersion));
      p += sprintf(p, "  </tr>\r\n");

      p += sprintf(p, "  <tr>\r\n");
      p += sprintf(p, "    <td>Negotiated version</td>\r\n");
      p += sprintf(p, "    <td>%s</td>\r\n", tlsGetVersionName(tlsContext->version));
      p += sprintf(p, "  </tr>\r\n");

      p += sprintf(p, "  <tr>\r\n");
      p += sprintf(p, "    <td>Cipher suite</td>\r\n");
      p += sprintf(p, "    <td>%s</td>\r\n", tlsGetCipherSuiteName(tlsContext->cipherSuite.identifier));
      p += sprintf(p, "  </tr>\r\n");

      p += sprintf(p, "  <tr>\r\n");
      p += sprintf(p, "    <td>Client random</td>\r\n");
      p += sprintf(p, "      <td>\r\n");
      p += sprintf(p, "        %s\r\n", dumpArray(buffer, (uint8_t *) &tlsContext->clientRandom, 32));
      p += sprintf(p, "      </td>\r\n");
      p += sprintf(p, "  </tr>\r\n");

      p += sprintf(p, "  <tr>\r\n");
      p += sprintf(p, "    <td>Server random</td>\r\n");
      p += sprintf(p, "      <td>\r\n");
      p += sprintf(p, "        %s\r\n", dumpArray(buffer, (uint8_t *) &tlsContext->serverRandom, 32));
      p += sprintf(p, "      </td>\r\n");
      p += sprintf(p, "  </tr>\r\n");

      p += sprintf(p, "  <tr>\r\n");
      p += sprintf(p, "    <td>Session ID</td>\r\n");
      p += sprintf(p, "      <td>\r\n");
      p += sprintf(p, "        %s\r\n", dumpArray(buffer, tlsContext->sessionId, tlsContext->sessionIdLen));
      p += sprintf(p, "      </td>\r\n");
      p += sprintf(p, "  </tr>\r\n");

      p += sprintf(p, "  </table>\r\n");
      p += sprintf(p, "</body>\r\n");
      p += sprintf(p, "</html>\r\n");

      //Dump HTTP response
      TRACE_INFO("HTTP response:\r\n");
      TRACE_INFO("%s\r\n", response);

      //Send response to the client
      error = tlsWrite(tlsContext, response, strlen(response), NULL, 0);
      //Any error to report?
      if(error)
         break;

      //End of exception handling block
   } while(0);

   //Terminate SSL session
   tlsFree(tlsContext);

   //Release previously allocated memory
   free(buffer);
   free(response);

   //Debug message
   TRACE_INFO("Closing connection...\r\n");
   //Close socket
   closesocket(clientSocket);

   //Debug message
   TRACE_INFO("Connection closed...\r\n");
   //Release semaphore
   osSemaphoreRelease(connectionSemaphore);

   //Kill ourselves
   osThreadTerminate(NULL);
}


/**
 * @brief Display the contents of an array
 * @param[out] buffer Output buffer where to format the resulting string
 * @param[in] data Pointer to the data array
 * @param[in] length Number of bytes in the array
 * @return Pointer to the output buffer
 **/

const char_t *dumpArray(char_t *buffer, const uint8_t *data, size_t length)
{
   size_t i;
   char_t *p;

   //Point to the output buffer
   p = buffer;
   //Properly terminate the string
   p[0] = '\0';

   //Process input data
   for(i = 0; i < length; i++)
   {
      //Beginning of a new line?
      if(i != 0 && (i % 16) == 0)
         p += sprintf(p, "\r\n        ");
      //Display current data byte
      p += sprintf(p, "%02X ", data[i]);
   }

   //Return a pointer the output string
   return buffer;
}


/**
 * @brief TLS send callback (I/O abstraction layer)
 * @param[in] socket Handle that identifies a socket
 * @param[in] data Pointer to a buffer containing the data to be transmitted
 * @param[in] length Number of data bytes to send
 * @param[out] written Number of bytes that have been transmitted
 * @param[in] flags Unused parameter
 * @return Error code
 **/

error_t tlsSendCallback(TlsIoHandle socket, const void *data,
   size_t length, size_t *written, uint_t flags)
{
   int_t n;
   error_t error;

   //Send data
   n = send((int) socket, data, length, 0);

   //Check return value
   if(n > 0)
   {
      //Total number of data that have been written
      *written = n;
      //Successful write operation
      error = NO_ERROR;
   }
   else
   {
      //Timeout error?
      if(n == BSD_ERROR_WOULDBLOCK || n == BSD_ERROR_TIMEOUT)
         error = ERROR_TIMEOUT;
      else
         error = ERROR_WRITE_FAILED;
   }

   //Return status code
   return error;
}


/**
 * @brief TLS receive callback (I/O abstraction layer)
 * @param[in] socket Handle that identifies a socket
 * @param[out] data Buffer where to store the incoming data
 * @param[in] size Maximum number of bytes that can be received
 * @param[out] received Number of bytes that have been received
 * @param[in] flags Unused parameter
 * @return Error code
 **/

error_t tlsReceiveCallback(TlsIoHandle socket, void *data,
   size_t size, size_t *received, uint_t flags)
{
   int_t n;
   error_t error;

   //Send data
   n = recv((int) socket, data, size, 0);

   //Check return value
   if(n > 0)
   {
      //Total number of data that have been received
      *received = n;
      //Successful write operation
      error = NO_ERROR;
   }
   else
   {
      //Timeout error?
      if(n == BSD_ERROR_WOULDBLOCK || n == BSD_ERROR_TIMEOUT)
         error = ERROR_TIMEOUT;
      else
         error = ERROR_READ_FAILED;
   }

   //Return status code
   return error;
}
